%{
	#include <stdio.h>
	#include <stdlib.h>
	#include <string.h>
	#include <malloc.h>
	#include "scanner.h"
	char *stringconst;
	int VarType;
	int n;
	char *keyword[] = {"ACCESS", "ALIAS", "ANY", "APPEND", "AS", "BASE",
                   "BINARY", "BYVAL", "CALL", "CALLS", "CASE", "CDEL",
                   "COMMON", "CONST", "DATA", "DECLARE", "DEF FN", "DEFDBL",
                   "DEFINT", "DEFLNG", "DEFSNG", "DEFSTR", "DIM", "DO",
                   "DOUBLE", "$DYNAMIC", "ELSE", "ELSEIF", "END", "EXIT",
                   "EXTERN", "FIELD", "FOR", "FUNCTION", "GOSUB", "GOTO", "IF", "SEG",
                   "$INCLUDE", "INTEGER", "IS", "LET", "LOCAL", "LONG",
                   "LOOP", "NEXT", "OPTION", "OUTPUT", "READ", "RESTORE",
                   "RESUME", "RETURN", "SELECT", "SHARED", "SIGNAL", "SINGLE", 
                   "STATIC", "$STATIC", "STEP", "STRING", "SUB", "THEN", "TO", "DEF",
                   "TYPE", "UNTIL", "USING", "WEND", "WHILE", "REDIM",
			 "START_C", "END_C", "START_ASM","END_ASM","EndKeywords"};
%}
digit	[0-9]
letter	[A-Za-z_]
suffix	[!&#$%]

%%
	/* Rules */
{digit}+	{
		VarType = VAR_INTEGER;
		return TK_NUMBER;
		}
"."{digit}+	{
		VarType = VAR_INTEGER;
		return TK_NUMBER;
		}
{digit}+{suffix}	{
		yytext[yyleng-1] = 0;
		VarType = VAR_INTEGER;
		return TK_NUMBER;
		}
{digit}+"."{digit}+ {
		VarType = VAR_SINGLE;
		return TK_NUMBER;
		}
{digit}+"."{digit}+{suffix}	{
		yytext[yyleng-1] = 0;
		VarType = VAR_SINGLE;
		return TK_NUMBER;
		}


	/* Comments */
"REM"	{
		int c = ' ';
		while(c != '\n')
			c = input();
		return TK_COMMENT;
	}
"'"	{
		int c = ' ';
		while(c != '\n')
			c = input();
		return TK_COMMENT;
	}

	/* Highest Important Operators */
"AND"	return TK_OPERATOR;
"OR"	return TK_OPERATOR;
"XOR"	return TK_OPERATOR;
"NOT"	return TK_OPERATOR;
"MOD"	return TK_OPERATOR;
"IMP"	return TK_OPERATOR;
"EQV"	return TK_OPERATOR;

	/* The dot between types and members */
"."	return TK_DOT;

"#"{letter}({letter}|{digit})*	{
				yytext = strdup(yytext+1);
				VarType = VAR_FILE;
				return TK_UNKNOWN;
				}
"#"{digit}*	{
				yytext = strdup(yytext+1);
				VarType = VAR_FILE;
				return TK_NUMBER;
				}

"&H"({letter}|{digit})*	{
				yytext = strdup(yytext+2);
				return TK_HEX;
				}

"&H"({letter}|{digit})*{suffix}	{
				yytext[yyleng-1] = 0;
				yytext = strdup(yytext+2);
				return TK_HEX;
				}

	/* Keyword (normal) */
{letter}({letter}|{digit})*	{
				n=0;
				while(strcmp(keyword[n],"EndKeywords") != 0)
				{
					if(strcmp(keyword[n],yytext) == 0)
						return TK_KEYWORD;
					n++;
				}

				/* Unknown Variable or Function */
				VarType = DefStruct->DefaultType;
				return TK_UNKNOWN;
				}

	/* Keyword (pre-processor) */
"$"{letter}({letter}|{digit})*	{
				n=0;
				while(strcmp(keyword[n],"EndKeywords") != 0)
				{
					if(strcmp(keyword[n],yytext) == 0)
						return TK_KEYWORD;
					n++;
				}
				}

	/* Variables */
{letter}({letter}|{digit})*"$"	{
				yytext[yyleng-1] = 0;
				VarType = VAR_STRING;
				return TK_UNKNOWN;
				}
{letter}({letter}|{digit})*"#"	{
				yytext[yyleng-1] = 0;
				VarType = VAR_DOUBLE;
				return TK_UNKNOWN;
				}
{letter}({letter}|{digit})*"!"	{
				yytext[yyleng-1] = 0;
				VarType = VAR_SINGLE;
				return TK_UNKNOWN;
				}
{letter}({letter}|{digit})*"&"	{
				yytext[yyleng-1] = 0;
				VarType = VAR_LONG;
				return TK_UNKNOWN;
				}
{letter}({letter}|{digit})*"%"	{
				yytext[yyleng-1] = 0;
				VarType = VAR_INTEGER;
				return TK_UNKNOWN;
				}

">="	return TK_OPERATOR;
"<="	return TK_OPERATOR;
"<>"	return TK_OPERATOR;
"++"	return TK_OPERATOR;
"--"	return TK_OPERATOR;
"+="	return TK_OPERATOR;
"-="	return TK_OPERATOR;
"/="	return TK_OPERATOR;
"*="	return TK_OPERATOR;

[-+/*=><\\^]	return TK_OPERATOR;

[ \t]	;
[\n]	return TK_NEWLINE;


\"[^"\n]*["\n]	{
			stringconst = strdup(yytext);
			if(stringconst[yyleng-1] != '"')
				return 254;
			return TK_STRINGCONSTANT;
			}

[().,;]	return TK_SYMBOL;

":"	return TK_SOFTLINE;

^({letter}|{digit})*":"$	return TK_LABEL;

<<EOF>> {
		
	return TK_EOF;
}

.	return TK_UNRECOGNIZE;

%%

int yywrap(void) {
	return 1;
}

char yyinput()
{
	return input();
}
